/*
 * Copyright (c) 2003 Marcus Glocker
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <dirent.h>

#include "../libmy/file.h"
#include "../libmy/str.h"

#include "proto.h"
#include "extern.h"

static const char behave[] = "Behave !\r\n";
static const char version[] = "bot 1.6 Marcus Glocker <marcus@nazgul.ch>";

static const struct {
	char	line[128];
} help[] = {
	{ "Channel Commands:" },
	{ "<single keyword>           Search keyword in database" },
	{ "!list                      List all available modules" },
	{ "!<module>                  Start a module" },
	{ "Private Commands:" },
	{ "quit                       Bot quits" },
	{ "help                       Display this help" },
	{ "info    <hostmask>         Show user info from userlist" },
	{ "+user   <hostmask>         Add user to userlist" },
	{ "-user   <hostmask>         Remove user from userlist" },
	{ "perm    <hostmask> <perm>  Change user permissions in userlist"},
	{ "join    #<channel>         Bot joins channel" },
	{ "part    #<channel>         Bot leaves channel" },
	{ "topic   #<channel> <topic> Set topic on channel" },
	{ "+op     #<channel> <nick>  Give nick operator mode on channel" },
	{ "-op     #<channel> <nick>  Remove nick operator mode on channel" },
	{ "kickban #<channel> <nick>  Kicks and bans nick on channel" }
};

int
irc_regpost(const int sfd)
{
	net_snd(sfd, "PASS secure\n");
	net_snd(sfd, "NICK %s\n", config.botnick);
	net_snd(sfd, "USER %s %s %s :%s\n",
	    config.botnick, config.hostname, config.ircserver, config.realname);

	return 0;
}

int
irc_regpre(const char *src, const int sfd)
{
	int	i = 0, r = 0;
	char	tmp[1024], ircserver[1024], channel[1024];

	strcutw(tmp, src, 1, sizeof(tmp));
	strcuts(ircserver, tmp, ':', ' ', sizeof(ircserver));
	strlcpy(config.ircserver, ircserver, sizeof(config.ircserver));

	r = irc_countwords(config.channel);

	for (i = 1; i <= r; i++) {
		strcutw(channel, config.channel, i, sizeof(channel));

		net_snd(sfd, "JOIN #%s\n", channel);
		net_snd(sfd, "MODE #%s +tn\n", channel);

		if (strcmp(config.topic, "0") != 0)
			net_snd(sfd, "TOPIC #%s :%s\n", channel, config.topic);

		net_snd(sfd, "PRIVMSG #%s :Logged on %s\n",
		    channel, config.ircserver);
	}

	return 0;
}

int
irc_listen(const char *src, const int sfd)
{
	int	botnet = 0, r = 0;
	char	tmp[1024];

	/* server actions */
	if (src[0] != ':') {
		/* ping, pong */
		if (strncmp(src, "PING", 4) == 0) {
			strcutw(tmp, src, 2, sizeof(tmp));
			net_snd(sfd, "PONG %s", tmp);
			r = 0;
		}
		/* error */
		if (strncmp(src, "ERROR", 5) == 0)
			r = 0;

		return r;
	}

	/* ignore myself */
	strcuts(tmp, src, ':', '!', sizeof(tmp));
	if (strcmp(tmp, config.botnick) == 0)
		return 1;

	/* ignore some botnet actions */
	strcuts(tmp, src, '!', ' ', sizeof(tmp));
	if (irc_perm(tmp, "b") == 0)
		botnet = 1;

	/* check the action */
	strcutw(tmp, src, 2, sizeof(tmp));

	if (strcmp(tmp, "001") == 0)
		r = 2;
	if (strcmp(tmp, "JOIN") == 0)
		r = 3;
	if (strcmp(tmp, "NOTICE") == 0)
		r = 4;
	if (strcmp(tmp, "PART") == 0)
		r = 5;
	if (strcmp(tmp, "QUIT") == 0)
		r = 6;
	if (strcmp(tmp, "MODE") == 0 && botnet != 1)
		r = 7;
	if (strcmp(tmp, "KICK") == 0 && botnet != 1)
		r = 8;
	if (strcmp(tmp, "PRIVMSG") == 0) {
		strcutw(tmp, src, 3, sizeof(tmp));
		if (tmp[0] == '#')
			r = 9;	/* channel msg */
		else
			r = 10;	/* private msg */
	}

	return r;
}

int
irc_notice(const char *src, const int sfd)
{
	char	hostmask[1024], nick[1024];

	/* ignore server notices */
	if (strcuts(nick, src, ':', '!', sizeof(nick)) == -1)
		return -1;

	strcuts(hostmask, src, '!', ' ', sizeof(hostmask));

	/* no use for notice yet */
	net_snd(sfd, "NOTICE %s :Received notice from you.\n", nick);
	
	return 0;
}

int
irc_join(const char *src, const int sfd)
{
	char	hostmask[1024], channel[1024], nick[1024], tmp[1024];

	strcuts(nick, src, ':', '!', sizeof(nick));
	strcuts(hostmask, src, '!', ' ', sizeof(hostmask));
	strcutw(tmp, src, 3, sizeof(tmp));
	strcuts(channel, tmp, ':', '\r', sizeof(channel));

	if (irc_perm(hostmask, "o") == 0)
		net_snd(sfd, "MODE %s +o %s\n", channel, nick);

	if (irc_perm(hostmask, "b") == 0)
		net_snd(sfd, "PRIVMSG %s :Welcome Robot !\n", channel);

	return 0;
}

int
irc_mode(const char *src, const int sfd)
{
	char	who_hostmask[1024];
	char	hostmask[1024], channel[1024], nick[1024], who[1024], m[1024];

	strcuts(nick, src, ':', '!', sizeof(nick));
	strcuts(hostmask, src, '!', ' ', sizeof(hostmask));
	strcutw(channel, src, 3, sizeof(channel));
	strcutw(m, src, 4, sizeof(m));
	strcutw(who, src, 5, sizeof(who)); 

	/* cut out \r\n */
	who[strlen(who)-2] = '\0';

	/* deop */
	if (strcmp(m, "-o") == 0) {
		/* deoped himself */
		if (strcmp(nick, who) == 0) {
			net_snd(sfd, "PRIVMSG %s :You don't like power, %s ?\n",
			    channel, nick);
			return 0;
		}

		/* get hostmask from deoped user */
		irc_who(who_hostmask, who, sfd, sizeof(who_hostmask));

		/* do we have deoped user in userlist */
		if (irc_perm(who_hostmask, "fdbm") == 0) {
			net_snd(sfd, "KICK %s %s :%s", channel, nick, behave);
			net_snd(sfd, "MODE %s +o %s\n", channel, who);
		}

		/* if deoped was a bot, ban attacker */
		if (irc_perm(who_hostmask, "b") == 0)
			net_snd(sfd, "MODE %s +b %s\n", channel, hostmask);
	}

	return 0;
}

int
irc_kick(const char *src, const int sfd)
{
	char	who_hostmask[1024];
	char	hostmask[1024], channel[1024], nick[1024], who[1024];

	strcuts(nick, src, ':', '!', sizeof(nick));
	strcuts(hostmask, src, '!', ' ', sizeof(hostmask));
	strcutw(channel, src, 3, sizeof(channel));
	strcutw(who, src, 4, sizeof(who));

	/* kicked himself */
	if (strcmp(nick, who) == 0) {
		net_snd(sfd, "PRIVMSG %s :%s seems to be masochistic !\n",
		    channel, nick);
		return 0;
	}

	/* get hostmask from kicked user */
	irc_who(who_hostmask, who, sfd, sizeof(who_hostmask));

	/* do we have kicked user in userlist */
	if (irc_perm(who_hostmask, "fdbm") == 0) {
		net_snd(sfd, "KICK %s %s :%s", channel, nick, behave);
		net_snd(sfd, "INVITE %s :%s\n", who, channel);
	}

	/* if kicked was a bot, ban attacker */
	if (irc_perm(who_hostmask, "b") == 0)
		net_snd(sfd, "MODE %s +b %s\n", channel, hostmask);

	return 0;
}

int
irc_privpub(const char *src, const int sfd)
{
	int	i;
	char	prgs[1024], prgf[1024], opt[1024];
	char	key[1024], lower[1024], tmp[1024], ch[1], buf[8192];
	char	hostmask[1024], channel[1024], nick[1024], cmd[1024];

	strcuts(nick, src, ':', '!', sizeof(nick));
	strcuts(hostmask, src, '!', ' ', sizeof(hostmask));
	strcutw(channel, src, 3, sizeof(channel));
	strcutw(cmd, src, 4, sizeof(cmd));
	strcutw(opt, src, 5, sizeof(opt));

	if (cmd[1] != '!' && cmd[strlen(cmd) - 2] == '\r') {
		strcuts(tmp, cmd, ':', '\r', sizeof(tmp));
		file_lower(lower, tmp, sizeof(lower));
		if (fparse(key, lower, config.filedb, sizeof(key)) != -1)
			net_snd(sfd, "PRIVMSG %s :%s\n", channel, key);
		return 0;
	}

	if (cmd[1] == '!' && strcmp(config.moddir, "0") == 0) {
		net_snd(sfd, "PRIVMSG %s :Modules are not activated\n",
		    channel);
		return 0;
	}

	if (strcmp(cmd, ":!list\r\n") == 0) {
		DIR		*odir;
		struct dirent	*rdir;

		if ((odir = opendir(config.moddir)) == NULL)
			return -1;
		net_snd(sfd, "PRIVMSG %s :%s modules:\n", channel,
		    config.botnick);
		while ((rdir = readdir(odir)) != NULL) {
			if (rdir->d_name[0] == '.' ||
			    strcmp(rdir->d_name, "modtab") == 0)
				continue;
			net_snd(sfd, "PRIVMSG %s :%s\n", channel, rdir->d_name);
		}
		closedir(odir);
		return 0;
	}

	if (cmd[1] == '!') {
		int	pid;
		int	fds1[2], fds2[2];

		pid = fork();
		if (pid > 0)
			return 0;

		if (cmd[strlen(cmd) - 2] == '\r') {
			strcuts(prgs, cmd, '!', '\r', sizeof(prgs));
			opt[0] = '\0';
		} else if (opt[strlen(opt) - 2] == '\r') {
			strcuts(prgs, cmd, '!', '\0', sizeof(prgs));
			opt[strlen(opt) - 2] = '\0';
		} else {
			return -1;
		}

		snprintf(prgf, sizeof(prgf), "%s/%s", config.moddir, prgs);

		flogd(config.filelog, "*** Execute module [%s]: %s %s\n", prgs,
		    prgf, opt);

		pipe(fds1);
		pipe(fds2);
		pid = fork();
		if (pid == 0) {
			close(fds1[1]);
			close(fds2[0]);
			dup2(fds1[0], STDIN_FILENO);
			dup2(fds2[1], STDOUT_FILENO);
			chdir(config.moddir);
			execl(prgf, prgs, opt, (char *)NULL);
			exit(0);
		}
		close(fds1[0]);
		close(fds2[1]);
		close(fds1[1]);

		i = 0;
		while (read(fds2[0], ch, 1) > 0 && i < sizeof(buf)) {
			if (ch[0] == '\n') {
				buf[i] = '\0';
				net_snd(sfd, "PRIVMSG %s :%s\n", channel, buf);
				i = 0;
				continue;
			}
			buf[i] = ch[0];
			i++;
		}
		close(fds2[0]);

		exit(0);
	}

	return 0;
}

int
irc_privprv(const char *src, const int sfd)
{
	int	i, r = 0;
	char	tmp[1024], who_hostmask[1024], perm[1024];
	char	cmdd[1024], tabs[128];
	char	hostmask[1024], cmd[1024], hc[1024], nick[1024];

	strcuts(nick, src, ':', '!', sizeof(nick));
	strcuts(hostmask, src, '!', ' ', sizeof(hostmask));
	strcutw(cmd, src, 4, sizeof(cmd));
	strcutw(hc, src, 5, sizeof(hc));

	/* ctcp version request */
	if (strcmp(cmd, ":\x01VERSION\x01\r\n") == 0) {
		net_snd(sfd, "NOTICE %s :\x01VERSION %s\x01\n", nick, version);
		return r;
	}

	/* quit */
	if (strcmp(cmd, ":quit\r\n") == 0) {
		if (irc_perm(hostmask, "m") == 0) {
			net_snd(sfd, "QUIT :Yes Master !\r\n");
			flogd(config.filelog, "*** bot stopped (pid %d)\n",
			    ppid);
			close(sfd);
			exit(0);
		} else {
			r = -1;
		}
	}

	/* help */
	if (strcmp(cmd, ":help\r\n") == 0) {
	 	for (i = 0; i < 15; i++)	
			net_snd(sfd, "PRIVMSG %s :%s\r\n", nick, help[i].line);
		net_snd(sfd, "PRIVMSG %s :Admin: %s\r\n", nick, config.admin);
		return r;
	}

	/* show user infos from userlist */
	if (strcmp(cmd, ":info") == 0) {
		hc[strlen(hc)-2] = '\0';
		if (fparse(tmp, hc, config.fileuser, sizeof(tmp))
		    == -1) {
			net_snd(sfd,
			"PRIVMSG %s :%s doesnt exist in userlist !\n",
			    nick, hc);
			return -1;
		}
		net_snd(sfd, "PRIVMSG %s :Hostmask: %s | Permissions: %s\n",
		    nick, hc, tmp);
		return r;
	}

	/* add user to userlist */
	if (strcmp(cmd, ":+user") == 0) {
		if (irc_perm(hostmask, "dbm") == 0) {
			hc[strlen(hc)-2] = '\0';
			/* user already exist */
			if (fparse(tmp, hc, config.fileuser, sizeof(tmp))
			    != -1) {
				net_snd(sfd,
				"PRIVMSG %s :%s already exist in userlist !\n",
				nick, hc);
				return -1;
			}
			file_tabs(tabs, hc, 40, sizeof(tabs));
			snprintf(tmp, sizeof(tmp), "%s%sn\n", hc, tabs);
			file_write(tmp, config.fileuser);
			net_snd(sfd, "PRIVMSG %s :%s added to userlist.\n",
			    nick, hc);
			return r;
		} else {
			r = -1;
		}
	}

	/* remove user from userlist */
	if (strcmp(cmd, ":-user") == 0) {
		if (irc_perm(hostmask, "dbm") == 0) {
			hc[strlen(hc)-2] = '\0';
			/* user doesnt exist */
			if (fparse(tmp, hc, config.fileuser, sizeof(tmp))
			    == -1) {
				net_snd(sfd,
				"PRIVMSG %s :%s doesnt exist in userlist !\n",
				nick, hc);
				return -1;
			}
			file_rmline(hc, config.fileuser);
			net_snd(sfd, "PRIVMSG %s :%s removed from userlist.\n",
			    nick, hc);
			return r;
		} else {
			r = -1;
		}
	}

	/* change user permissions */
	if (strcmp(cmd, ":perm") == 0) {
		if (irc_perm(hostmask, "dbm") == 0) {
			strcutw(perm, src, 6, sizeof(perm));
			perm[strlen(perm)-2] = '\0';
			if (fparse(tmp, hc, config.fileuser, sizeof(tmp))
			    == -1) {
				net_snd(sfd,
				"PRIVMSG %s :%s doenst exist in userlist !\n",
				nick, hc);
				return -1;
			}
			file_tabs(tabs, hc, 40, sizeof(tabs));
			snprintf(tmp, sizeof(tmp), "%s%s%s", hc, tabs, perm);
			file_chline(hc, tmp, config.fileuser);
			net_snd(sfd,
			    "PRIVMSG %s :%s permissions changed to %s.\n",
			    nick, hc, perm);
			return r;
		} else {
			r = -1;
		}
	}

	/* commands from here on, needs a channel */
	if (hc[0] != '#')
		goto error;

	/* join other channel */
	if (strcmp(cmd, ":join") == 0) {
		if (hc[strlen(hc)-1] != '\n') {
			net_snd(sfd, "PRIVMSG %s :Wrong channel syntax!\n",
			    nick);
			return -1;
		}
		if (irc_perm(hostmask, "m") == 0) {
			net_snd(sfd, "PRIVMSG %s :Joining channel %s\n",
			    nick, hc);
			net_snd(sfd, "JOIN %s\n", hc);
			return r;
		} else {
			r = -1;
		}
	}

	/* part channel */
	if (strcmp(cmd, ":part") == 0) {
		if (hc[strlen(hc)-1] != '\n') {
			net_snd(sfd, "PRIVMSG %s :Wrong channel syntax!\n",
			    nick);
			return -1;
		}
		if (irc_perm(hostmask, "m") == 0) {
			net_snd(sfd, "PRIVMSG %s :Leaving channel %s\n",
			    nick, hc);
			net_snd(sfd, "PART %s\n", hc);
			return r;
		} else {
			r = -1;
		}
	}

	/* set topic */
	if (strcmp(cmd, ":topic") == 0) {
		if (irc_perm(hostmask, "fdbm") == 0) {
			irc_cutcmddata(cmdd, src, sizeof(cmdd));
			net_snd(sfd, "TOPIC %s :%s\n", hc, cmdd);
			return r;
		} else {
			r = -1;
		}
	}

	/* op */
	if (strcmp(cmd, ":+op") == 0) {
		if (irc_perm(hostmask, "dbm") == 0) {
			strcutw(cmdd, src, 6, sizeof(cmdd));
			net_snd(sfd, "MODE %s +o %s", hc, cmdd);
			return r;
		} else {
			r = -1;
		}
	}

	/* deop */
	if (strcmp(cmd, ":-op") == 0) {
		if (irc_perm(hostmask, "dbm") == 0) {
			strcutw(cmdd, src, 6, sizeof(cmdd));
			net_snd(sfd, "MODE %s -o %s", hc, cmdd);
			return r;
		} else {
			r = -1;
		}
	}
	
	/* kickban */
	if (strcmp(cmd, ":kickban") == 0) {
		if (irc_perm(hostmask, "dbm") == 0) {
			strcutw(cmdd, src, 6, sizeof(cmdd));
			/* cut out \r\n */
			cmdd[strlen(cmdd)-2] = '\0';
			irc_who(who_hostmask, cmdd, sfd, sizeof(who_hostmask));
			net_snd(sfd, "KICK %s %s :Stay away !\n", hc, cmdd);
			net_snd(sfd, "MODE %s +b %s\n", hc, who_hostmask);
			return r;
		} else {
			r = -1;
		}
	}

	/* no permissions or command not found */
	error:
	if (r == -1)
		net_snd(sfd, "PRIVMSG %s :Permission denied !\r\n", nick);
	else
		net_snd(sfd, "PRIVMSG %s :Excuse me ?\r\n", nick);

	return r;
}

int
irc_who(char *dst, const char *src, const int sfd, const int dsize)
{
	char	in[5120], nick[1024], domain[1024], hostmask[1024];

	net_snd(sfd, "WHO %s\n", src);

	recv(sfd, in, sizeof(in), 0);
	flogd(config.filelog, in);

	strcutw(nick, in, 5, sizeof(nick));
	strcutw(domain, in, 6, sizeof(domain));

	snprintf(hostmask, sizeof(hostmask), "%s@%s", nick, domain);

	strlcpy(dst, hostmask, dsize);

	return 0;
}

int
irc_cutcmddata(char *dst, const char *src, const int dsize)
{
	int	i, j;

	for (i = 0, j = 0; j < 5; i++) {
		if (src[i] == ' ')
			j++;
	}

	for (j = 0; src[i] != '\r' && j != dsize-1; i++, j++)
		dst[j] = src[i];

	/* terminate dst */
	dst[j] = '\0';

	return 0;
}

int
irc_countwords(const char *src)
{
	int	i, len, words = 1;

	if ((len = strlen(src)) == 0)
		return -1;

	for (i = 0; i < len; i++) {
		if (src[i] == ' ')
			words++;
	}

	return words;
}

int
irc_perm(const char *hostmask, const char *perm)
{
	int	i, j = 0;
	char	tmp[1024];

	if (fparse(tmp, hostmask, config.fileuser, sizeof(tmp)) == -1)
		return -1;

	while (j < strlen(perm)) {
		for (i = 0; i < strlen(tmp); i++) {
			if (tmp[i] == perm[j])
				return 0;
		}
		j++;
	}

	return -1;
}
