/*
 * Copyright (c) 2004 Marcus Glocker
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

/*
 * undeadly module for bot
 *
 * This module connects the OpenBSD Journal at http://undeadly.org and scans
 * the main page for the latest headlines.
 *
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>

#define IS 102400	/* max size in bytes received from main page */
#define HL 3		/* number of headlines displayed */

const static char
	*request = "GET /cgi?action=rss HTTP/1.1\r\nHost: undeadly.org\r\n\r\n";
const static char
	*p_link  = "<link>";
const static char
	*p_title = "<title>";
const static char
	*news_file = ".undeadly.news";

void 	usage(void);
char *	format(const char *);

char *
format(const char *title)
{
	int	i;
	char	esc[16];
	char	*p, *save;

	if ((p = malloc(1024)) == NULL) {
		perror("malloc");
		exit(1);
	}
	save = p;
	memset(p, 0, 1024);

	while (*title != '>')
		title++;

	title++;
	while (*title != '<') {
		if (*title == '&') {
			i = 0;
			title++;
			while (*title != ';') {
				esc[i] = *title;
				i++; title++;
			}
			esc[i] = '\0';
			if (strcmp(esc, "amp") == 0)
				*p = '&';
		} else {
			*p = *title;
		}
		title++;
		p++;
	}

	return save;
}

void
usage(void)
{
	extern char	*__progname;

	printf("usage: %s ", __progname);
	printf("[-hfs]\n\n");
	printf("options:\n");
	printf("  -h (help)\t: This help.\n");
	printf("  -f (force)\t: Show headlines no matter if new.\n");
	printf("  -s (silent)\t: Show headlines if new, without comments.\n");
}

int
main(int argc, char *argv[])
{
	FILE			*file = NULL;
	int			sfd, i, j, r, news, ignore, fd, ch, opt;
	char			receive[IS], index[IS], buf[1024], c[1];
	char			title[1024], link[1024], msg1[8192], msg2[8192];
	struct hostent		*host;
	struct sockaddr_in	sa;
	struct stat		st;

	opt = 0;
	while ((ch = getopt(argc, argv, "hfs")) != -1) {
		switch (ch) {
			case 'h':
				usage();
				exit(0);
				break;
			case 'f':
				opt = 1;
				break;
			case 's':
				opt = 2;
				break;
			default:
				break;
		}
	}

	if ((host = gethostbyname("undeadly.org")) == NULL) {
		perror("gethostbyname");
		exit(1);
	}

	if ((sfd = socket(AF_INET, SOCK_STREAM, 0))  == -1) {
		perror("socket");
		exit(1);
	}

	sa.sin_family = AF_INET;
	sa.sin_port = htons(80);
	sa.sin_addr = *((struct in_addr *) host->h_addr);

	memset(&(sa.sin_zero), 0, sizeof(sa.sin_zero));

	if (connect(sfd, (struct sockaddr *) &sa, sizeof(struct sockaddr))
	    == -1) {
		perror("connect");
		exit(1);
	}

	if (send(sfd, request, strlen(request), 0) == -1) {
		perror("send");
		exit(1);
	}

	memset(index, 0, sizeof(index));
	memset(receive, 0, sizeof(receive));
	while ((r = recv(sfd, receive, sizeof(receive), 0)) > 0) {
		strlcat(index, receive, sizeof(index));
		memset(receive, 0, sizeof(receive));
	}

	if (index[0] == '\0')
		exit(0);

	memset(msg1, 0, sizeof(msg1));
	i = 0, j = 0, news = 0, ignore = 0;
	while (index[i] != '\0' && j < sizeof(buf) && news < HL) {
		if (index[i] == '\n') {
			if (strstr(buf, p_title) != NULL) {
				if (ignore > 3) {
					snprintf(title, sizeof(title),
					    "[OpenBSD Journal] %s - ",
					    format(buf));
				} else {
					ignore++;
				}
			}
			if (strstr(buf, p_link) != NULL) {
				if (ignore > 3) {
					snprintf(link, sizeof(link),
					    "%s\n", format(buf));
					strlcat(msg1, title, sizeof(msg1));
					strlcat(msg1, link, sizeof(msg1));
					news++;
				} else {
					ignore++;
				}
			}
			memset(buf, 0, sizeof(buf));
			i++; j = 0;
		}
		buf[j] = index[i];
		i++; j++;
	}

	close(sfd);

	if (argc > 1 && strcmp(argv[1], "force") == 0) {
		printf("%s", msg1);
		return 0;
	}

	if (stat(news_file, &st) == -1) {
		file = fopen(news_file, "w");
		fprintf(file, "%s", msg1);
		fclose(file);
		printf("%s", msg1);
	} else {
		memset(msg2, 0, sizeof(msg2));
		i = 0;
		fd = open(news_file, O_RDONLY);
		while (read(fd, c, 1) > 0) {
			msg2[i] = c[0];
			i++;
		}
		close(fd);

		if (strcmp(msg1, msg2) == 0) {
			if (opt == 0)
				printf("No News.\n");
			if (opt == 1)
				printf("%s", msg1);
		} else {
			file = fopen(news_file, "w");
			fprintf(file, "%s", msg1);
			fclose(file);
			printf("%s", msg1);
		}
	}

	return 0;
}
